use std::fs;

use anyhow::Error;
use envhub_ext::{rtx::Rtx, Extension};
use envhub_hm::switch::switch_env;
use envhub_providers::{github::Github, local::Local, s3::S3, Provider};
use envhub_stow::stow::stow;
use owo_colors::OwoColorize;

use crate::helpers::{copy_home_nix, get_home_manager_dir, install_packages, read_envhub_file};

pub fn use_environment(name: &str) -> Result<(), Error> {
    let scheme = match name.split(":").collect::<Vec<&str>>().len() > 1 {
        false => "local",
        true => name.split(":").collect::<Vec<&str>>()[0],
    };
    let source: Box<dyn Provider> = match scheme {
        "github" => Box::new(Github::new()),
        "local" => Box::new(Local::new()),
        "s3" => Box::new(S3::new()),
        _ => panic!("Unknown scheme: {}", scheme),
    };
    let name = match name.split(":").collect::<Vec<&str>>().len() > 1 {
        false => name,
        true => name.split(":").collect::<Vec<&str>>()[1],
    };

    let path = match scheme {
        "local" => fs::canonicalize(name)?.to_str().unwrap().to_string(),
        _ => format!("{}:{}", scheme, name),
    };

    source.load(name)?;

    let home_manager_dir = get_home_manager_dir(scheme, name)?;

    copy_home_nix(&home_manager_dir)?;

    let config = read_envhub_file(&home_manager_dir)?;

    if config.package_manager.is_some() && config.package_manager != Some("nix".into()) {
        install_packages(&config)?;
    }

    if config.rtx.is_some() {
        let rtx: Box<dyn Extension> = Box::new(Rtx::new());
        rtx.load(&config)?;
    }

    switch_env(Some(&home_manager_dir), &config)?;

    let state = toml::to_string(&config)?;

    if config.symlink_manager == Some("stow".into()) {
        let target = std::env::var("HOME")?;
        let package = "dotfiles";
        stow(&home_manager_dir, &target, &package)?;
    }

    fs::write(
        format!("{}/.envhub/current", std::env::var("HOME")?),
        format!(
            "{}\n{}\n{}",
            &home_manager_dir,
            &config.symlink_manager.unwrap_or("home-manager".into()),
            &path
        ),
    )?;

    fs::write(
        format!("{}/.envhub/envhub.toml", std::env::var("HOME")?),
        format!(
            "# This file is automatically @generated by envhub.\n# It is not intended for manual editing.\n\n{}",
            state
        ),
    )?;

    println!(
        "{} Successfully used environment: {}",
        "[✓]".bright_green(),
        home_manager_dir.bright_green()
    );

    Ok(())
}
